@@ -30,7 +30,7 @@ from __future__ import generators
 import base64, errno, re
 try: import posix1e
 except ImportError: pass
-import static, Globals, metadata, connection, rorpiter, log, C, \
+import static, Globals, eas_acls, connection, metadata, rorpiter, log, C, \
 	   rpath, user_group
 
 # When an ACL gets dropped, put name in dropped_acl_names.  This is
@@ -56,43 +56,71 @@ class ExtendedAttributes:
 
 	def read_from_rp(self, rp):
 		"""Set the extended attributes from an rpath"""
-		try: attr_list = rp.conn.xattr.listxattr(rp.path)
+		try:
+			attr_list = rp.conn.xattr.listxattr(rp.path, rp.issym())
 		except IOError, exc:
-			if exc[0] == errno.EOPNOTSUPP: return # if not sup, consider empty
-			if exc[0] == errno.EACCES:
-				log.Log("Warning: listattr(%s): %s" % (rp.path, exc), 3)
+			if exc[0] in (errno.EOPNOTSUPP, errno.EPERM, errno.ETXTBSY):
+				return # if not supported, consider empty
+			if exc[0] in (errno.EACCES, errno.ENOENT, errno.ELOOP):
+				log.Log("Warning: listattr(%s): %s" % (repr(rp.path), exc), 4)
 				return
 			raise
 		for attr in attr_list:
-			if not attr.startswith('user.'):
-				# Only preserve user extended attributes
+			if attr.startswith('system.'):
+				# Do not preserve system extended attributes
 				continue
-			try: self.attr_dict[attr] = rp.conn.xattr.getxattr(rp.path, attr)
+			if not rp.isdir() and attr == 'com.apple.ResourceFork':
+				# Resource Fork handled elsewhere, except for directories
+				continue
+			try:
+				self.attr_dict[attr] = \
+					rp.conn.xattr.getxattr(rp.path, attr, rp.issym())
 			except IOError, exc:
 				# File probably modified while reading, just continue
 				if exc[0] == errno.ENODATA: continue
 				elif exc[0] == errno.ENOENT: break
+				# Handle bug in pyxattr < 0.2.2
+				elif exc[0] == errno.ERANGE: continue
 				else: raise
 
 	def clear_rp(self, rp):
 		"""Delete all the extended attributes in rpath"""
-		for name in rp.conn.xattr.listxattr(rp.path):
-			try:
-				rp.conn.xattr.removexattr(rp.path, name)
-			except IOError, exc:
-				# SELinux attributes cannot be removed, and we don't want
-				# to bail out or be too noisy at low log levels.
-				if exc[0] == errno.EACCES:
-					log.Log("Warning: unable to remove xattr %s from %s"
-						% (name, rp.path), 7)
-					continue
-				else: raise
+		try:
+			for name in rp.conn.xattr.listxattr(rp.path, rp.issym()):
+				try:
+					rp.conn.xattr.removexattr(rp.path, name, rp.issym())
+				except IOError, exc:
+					# SELinux attributes cannot be removed, and we don't want
+					# to bail out or be too noisy at low log levels.
+					if exc[0] == errno.EACCES:
+						log.Log("Warning: unable to remove xattr %s from %s"
+							% (name, repr(rp.path)), 7)
+						continue
+					else: raise
+		except IOError, exc:
+			if exc[0] == errno.EOPNOTSUPP or exc[0] == errno.EPERM:
+				return # if not supported, consider empty
+			elif exc[0] == errno.ENOENT: # path is bad
+				log.Log("Warning: unable to clear xattrs on %s: %s" %
+						(repr(rp.path), exc), 3)
+				return
+			else: raise
 
 	def write_to_rp(self, rp):
 		"""Write extended attributes to rpath rp"""
 		self.clear_rp(rp)
 		for (name, value) in self.attr_dict.iteritems():
-			rp.conn.xattr.setxattr(rp.path, name, value)
+			try:
+				rp.conn.xattr.setxattr(rp.path, name, value, 0, rp.issym())
+			except IOError, exc:
+				# Mac and Linux attributes have different namespaces, so
+				# fail gracefully if can't call setxattr
+				if exc[0] in (errno.EOPNOTSUPP, errno.EPERM, errno.EACCES,
+						errno.ENOENT, errno.EINVAL):
+					log.Log("Warning: unable to write xattr %s to %s"
+							% (name, repr(rp.path)), 6)
+					continue
+				else: raise
 
 	def get(self, name):
 		"""Return attribute attached to given name"""
@@ -126,7 +154,11 @@ def EA2Record(ea):
 		if not val: str_list.append(name)
 		else:
 			encoded_val = base64.encodestring(val).replace('\n', '')
-			str_list.append('%s=0s%s' % (C.acl_quote(name), encoded_val))
+			try:
+				str_list.append('%s=0s%s' % (C.acl_quote(name), encoded_val))
+			except UnicodeEncodeError:
+				log.Log("Warning: unable to store Unicode extended attribute %s"
+							% repr(name), 3)
 	return '\n'.join(str_list)+'\n'
 
 def Record2EA(record):
@@ -170,30 +202,19 @@ class ExtendedAttributesFile(metadata.FlatFile):
 	_extractor = EAExtractor
 	_object_to_record = staticmethod(EA2Record)
 
-	def join(cls, rorp_iter, rbdir, time, restrict_index):
-		"""Add extended attribute information to existing rorp_iter"""
-		def helper(rorp_iter, ea_iter):
-			"""Add EA information in ea_iter to rorp_iter"""
-			collated = rorpiter.CollateIterators(rorp_iter, ea_iter)
-			for rorp, ea in collated:
-				assert rorp, (rorp, (ea.index, ea.attr_dict), time)
-				if not ea: ea = ExtendedAttributes(rorp.index)
-				rorp.set_ea(ea)
-				yield rorp
-			
-		ea_iter = cls.get_objects_at_time(rbdir, time, restrict_index)
-		if not ea_iter:
-			log.Log("Warning: Extended attributes file not found", 2)
-			ea_iter = iter([])
-		return helper(rorp_iter, ea_iter)
-
-static.MakeClass(ExtendedAttributesFile)
+def join_ea_iter(rorp_iter, ea_iter):
+	"""Update a rorp iter by adding the information from ea_iter"""
+	for rorp, ea in rorpiter.CollateIterators(rorp_iter, ea_iter):
+		assert rorp, "Missing rorp for index %s" % (ea.index,)
+		if not ea: ea = ExtendedAttributes(rorp.index)
+		rorp.set_ea(ea)
+		yield rorp
 
 
 class AccessControlLists:
 	"""Hold a file's access control list information
 
-	Since posix1e.ACL objects cannot be picked, and because they lack
+	Since posix1e.ACL objects cannot be pickled, and because they lack
 	user/group name information, store everything in self.entry_list
 	and self.default_entry_list.
 
@@ -208,7 +229,7 @@ class AccessControlLists:
 		"""Set self.entry_list and self.default_entry_list from text"""
 		self.entry_list, self.default_entry_list = [], []
 		for line in text.split('\n'):
-			comment_pos = text.find('#')
+			comment_pos = line.find('#')
 			if comment_pos >= 0: line = line[:comment_pos]
 			line = line.strip()
 			if not line: continue
@@ -220,6 +241,7 @@ class AccessControlLists:
 
 	def __str__(self):
 		"""Return text version of acls"""
+		if not self.entry_list: return ""
 		slist = map(self.entrytuple_to_text, self.entry_list)
 		if self.default_entry_list:
 			slist.extend(map(lambda e: "default:" + self.entrytuple_to_text(e),
@@ -355,7 +377,15 @@ def set_rp_acl(rp, entry_list = None, default_entry_list = None,
 	assert rp.conn is Globals.local_connection
 	if entry_list: acl = list_to_acl(entry_list, map_names)
 	else: acl = posix1e.ACL()
-	acl.applyto(rp.path)
+
+	try:
+		acl.applyto(rp.path)
+	except IOError, exc:
+		if exc[0] == errno.EOPNOTSUPP:
+			log.Log("Warning: unable to set ACL on %s: %s" % 
+					(repr(rp.path), exc), 4)
+			return
+		else: raise
 
 	if rp.isdir():
 		if default_entry_list:
@@ -368,12 +398,22 @@ def get_acl_lists_from_rp(rp):
 	assert rp.conn is Globals.local_connection
 	try: acl = posix1e.ACL(file=rp.path)
 	except IOError, exc:
-		if exc[0] == errno.EOPNOTSUPP: acl = None
+		if exc[0] == errno.EOPNOTSUPP:
+			acl = None
+		elif exc[0] == errno.ENOENT:
+			log.Log("Warning: unable to read ACL from %s: %s"
+					% (repr(rp.path), exc), 3)
+			acl = None
 		else: raise
 	if rp.isdir():
 		try: def_acl = posix1e.ACL(filedef=rp.path)
 		except IOError, exc:
-			if exc[0] == errno.EOPNOTSUPP: def_acl = None
+			if exc[0] == errno.EOPNOTSUPP:
+				def_acl = None
+			elif exc[0] == errno.ENOENT:
+				log.Log("Warning: unable to read default ACL from %s: %s"
+					% (repr(rp.path), exc), 3)
+				def_acl = None
 			else: raise
 	else: def_acl = None
 	return (acl and acl_to_list(acl), def_acl and acl_to_list(def_acl))
@@ -458,24 +498,15 @@ def list_to_acl(entry_list, map_names = 1):
 				"trigger further warnings" % (name,), 2)
 		dropped_acl_names[name] = name
 
-	def map_id_name(owner_pair, group = None):
-		"""Return id of mapped id and user given original owner_pair"""
-		id, name = owner_pair
-		Map = group and user_group.GroupMap or user_group.UserMap
-		if name: return Map.get_id_from_name(name)
-		else:
-			assert id is not None
-			return Map.get_id_from_id(id)
-
 	acl = posix1e.ACL()
 	for typechar, owner_pair, perms in entry_list:
 		id = None
 		if owner_pair:
 			if map_names:
-				if typechar == "u": id = map_id_name(owner_pair, 0)
+				if typechar == "u": id = user_group.acl_user_map(*owner_pair)
 				else:
 					assert typechar == "g", (typechar, owner_pair, perms)
-					id = map_id_name(owner_pair, 1)
+					id = user_group.acl_group_map(*owner_pair)
 				if id is None:
 					warn_drop(owner_pair[1])
 					continue
@@ -530,49 +561,14 @@ class AccessControlListFile(metadata.FlatFile):
 	_extractor = ACLExtractor
 	_object_to_record = staticmethod(ACL2Record)
 
-	def join(cls, rorp_iter, rbdir, time, restrict_index):
-		"""Add access control list information to existing rorp_iter"""
-		def helper(rorp_iter, acl_iter):
-			"""Add ACL information in acl_iter to rorp_iter"""
-			collated = rorpiter.CollateIterators(rorp_iter, acl_iter)
-			for rorp, acl in collated:
-				assert rorp, "Missing rorp for index %s" % (acl.index,)
-				if not acl: acl = AccessControlLists(rorp.index)
-				rorp.set_acl(acl)
-				yield rorp
-
-		acl_iter = cls.get_objects_at_time(rbdir, time, restrict_index)
-		if not acl_iter:
-			log.Log("Warning: Access Control List file not found", 2)
-			acl_iter = iter([])
-		return helper(rorp_iter, acl_iter)
-
-static.MakeClass(AccessControlListFile)
-
-
-def GetCombinedMetadataIter(rbdir, time, restrict_index = None,
-							acls = None, eas = None):
-	"""Return iterator of rorps from metadata and related files
-
-	None will be returned if the metadata file is absent.  If acls or
-	eas is true, access control list or extended attribute information
-	will be added.
-
-	"""
-	metadata_iter = metadata.MetadataFile.get_objects_at_time(
-		rbdir, time, restrict_index)
-	if not metadata_iter:
-		log.Log("Warning, metadata file not found.\n"
-				"Metadata will be read from filesystem.", 2)
-		return None
-	if eas:
-		metadata_iter = ExtendedAttributesFile.join(
-			metadata_iter, rbdir, time, restrict_index)
-	if acls:
-		metadata_iter = AccessControlListFile.join(
-			metadata_iter, rbdir, time, restrict_index)
-	return metadata_iter
-
+def join_acl_iter(rorp_iter, acl_iter):
+	"""Update a rorp iter by adding the information from acl_iter"""
+	for rorp, acl in rorpiter.CollateIterators(rorp_iter, acl_iter):
+		assert rorp, "Missing rorp for index %s" % (acl.index,)
+		if not acl: acl = AccessControlLists(rorp.index)
+		rorp.set_acl(acl)
+		yield rorp
+	
 
 def rpath_acl_get(rp):
 	"""Get acls of given rpath rp.
@@ -585,6 +581,11 @@ def rpath_acl_get(rp):
 	return acl
 rpath.acl_get = rpath_acl_get
 
+def rpath_get_blank_acl(index):
+	"""Get a blank AccessControlLists object (override rpath function)"""
+	return AccessControlLists(index)
+rpath.get_blank_acl = rpath_get_blank_acl
+
 def rpath_ea_get(rp):
 	"""Get extended attributes of given rpath
 
@@ -592,6 +593,13 @@ def rpath_ea_get(rp):
 
 	"""
 	ea = ExtendedAttributes(rp.index)
-	if not rp.issym(): ea.read_from_rp(rp)
+	if not rp.issock() and not rp.isfifo():
+		ea.read_from_rp(rp)
 	return ea
 rpath.ea_get = rpath_ea_get
+
+def rpath_get_blank_ea(index):
+	"""Get a blank ExtendedAttributes object (override rpath function)"""
+	return ExtendedAttributes(index)
+rpath.get_blank_ea = rpath_get_blank_ea
+
