@@ -1,4 +1,4 @@
-# Copyright 2002, 2003, 2004 Ben Escoto
+# Copyright 2002, 2003, 2004, 2005 Ben Escoto
 #
 # This file is part of rdiff-backup.
 #
@@ -21,15 +21,7 @@
 
 from __future__ import generators
 import tempfile, os, cStringIO
-import Globals, Time, Rdiff, Hardlink, rorpiter, selection, rpath, \
-	   log, static, robust, metadata, statistics, TempFile, eas_acls
-
-
-# This will be set to the time of the current mirror
-_mirror_time = None
-# This will be set to the exact time to restore to (not restore_to_time)
-_rest_time = None
-
+import static, rorpiter, FilenameMapping
 
 class RestoreError(Exception): pass
 
@@ -48,14 +40,17 @@ def Restore(mirror_rp, inc_rpath, target, restore_to_time):
 def get_inclist(inc_rpath):
 	"""Returns increments with given base"""
 	dirname, basename = inc_rpath.dirsplit()
+	if Globals.chars_to_quote: basename = FilenameMapping.unquote(basename)
 	parent_dir = inc_rpath.__class__(inc_rpath.conn, dirname, ())
 	if not parent_dir.isdir(): return [] # inc directory not created yet
 	index = inc_rpath.index
 
 	inc_list = []
 	for filename in parent_dir.listdir():
-		inc = parent_dir.append(filename)
-		if inc.isincfile() and inc.getincbase_str() == basename:
+		inc_info = rpath.get_incfile_info(filename)
+		if inc_info and inc_info[3] == basename:
+			inc = parent_dir.append(filename)
+			assert inc.isincfile()
 			inc_list.append(inc)
 	return inc_list
 
@@ -72,8 +67,8 @@ def ListChangedSince(mirror_rp, inc_rp, restore_to_time):
 	MirrorStruct.set_mirror_and_rest_times(restore_to_time)
 	MirrorStruct.initialize_rf_cache(mirror_rp, inc_rp)
 
-	old_iter = MirrorStruct.get_mirror_rorp_iter(_rest_time, 1)
-	cur_iter = MirrorStruct.get_mirror_rorp_iter(_mirror_time, 1)
+	old_iter = MirrorStruct.get_mirror_rorp_iter(MirrorStruct._rest_time, 1)
+	cur_iter = MirrorStruct.get_mirror_rorp_iter(MirrorStruct._mirror_time, 1)
 	collated = rorpiter.Collate2Iters(old_iter, cur_iter)
 	for old_rorp, cur_rorp in collated:
 		if not old_rorp: change = "new"
@@ -94,43 +89,23 @@ def ListAtTime(mirror_rp, inc_rp, time):
 	assert mirror_rp.conn is Globals.local_connection, "Run locally only"
 	MirrorStruct.set_mirror_and_rest_times(time)
 	MirrorStruct.initialize_rf_cache(mirror_rp, inc_rp)
-	old_iter = MirrorStruct.get_mirror_rorp_iter(_rest_time, 1)
+	old_iter = MirrorStruct.get_mirror_rorp_iter()
 	for rorp in old_iter: yield rorp
 	
-def Compare(src_iter, mirror_rp, inc_rp, compare_time):
-	"""Compares metadata in src_rp dir with metadata in mirror_rp at time"""
-	MirrorStruct.set_mirror_and_rest_times(compare_time)
-	MirrorStruct.initialize_rf_cache(mirror_rp, inc_rp)
-
-	mir_iter = MirrorStruct.get_mirror_rorp_iter(_rest_time, 1)
-	collated = rorpiter.Collate2Iters(src_iter, mir_iter)
-	changed_files_found = 0
-	for src_rorp, mir_rorp in collated: 
-		if not mir_rorp: change = "new"
-		elif not src_rorp: change = "deleted"
-		elif src_rorp == mir_rorp: continue
-		else: change = "changed"
-		changed_files_found = 1
-		path_desc = (src_rorp and src_rorp.get_indexpath() or
-					 mir_rorp.get_indexpath())
-		log.Log("%-7s %s" % (change, path_desc), 2)
-		if change == "changed": # Log more description of difference
-			assert not src_rorp.equal_verbose_auto(mir_rorp, 3)
-
-	if not changed_files_found:
-		log.Log("No changes found.  Directory matches archive data.", 2)
-	MirrorStruct.close_rf_cache()
-	return changed_files_found
-
 
 class MirrorStruct:
 	"""Hold functions to be run on the mirror side"""
-	_select = None # If selection command line arguments given, use Select here
+	# If selection command line arguments given, use Select here
+	_select = None
+	# This will be set to the time of the current mirror
+	_mirror_time = None
+	# This will be set to the exact time to restore to (not restore_to_time)
+	_rest_time = None
+	
 	def set_mirror_and_rest_times(cls, restore_to_time):
-		"""Set global variabels _mirror_time and _rest_time on mirror conn"""
-		global _mirror_time, _rest_time
-		_mirror_time = cls.get_mirror_time()
-		_rest_time = cls.get_rest_time(restore_to_time)
+		"""Set class variabels _mirror_time and _rest_time on mirror conn"""
+		MirrorStruct._mirror_time = cls.get_mirror_time()
+		MirrorStruct._rest_time = cls.get_rest_time(restore_to_time)
 
 	def get_mirror_time(cls):
 		"""Return time (in seconds) of latest mirror"""
@@ -169,8 +144,8 @@ class MirrorStruct:
 
 		"""
 		# use dictionary to remove dups
-		if not _mirror_time: d = {cls.get_mirror_time(): None}
-		else: d = {_mirror_time: None}
+		if not cls._mirror_time: d = {cls.get_mirror_time(): None}
+		else: d = {cls._mirror_time: None}
 		if not rp or not rp.index: rp = Globals.rbdir.append("increments")
 		for inc in get_inclist(rp): d[inc.getinctime()] = None
 		for inc in get_inclist(Globals.rbdir.append("mirror_metadata")):
@@ -201,11 +176,11 @@ class MirrorStruct:
 		unwanted files from the metadata_iter.
 
 		"""
-		if rest_time is None: rest_time = _rest_time
+		if rest_time is None: rest_time = cls._rest_time
 
-		rorp_iter = eas_acls.GetCombinedMetadataIter(
-			Globals.rbdir, rest_time, restrict_index = cls.mirror_base.index,
-			acls = Globals.acls_active, eas = Globals.eas_active)
+		metadata.SetManager()
+		rorp_iter = metadata.ManagerObj.GetAtTime(rest_time,
+												  cls.mirror_base.index)
 		if not rorp_iter:
 			if require_metadata:
 				log.Log.FatalError("Mirror metadata not found")
@@ -282,7 +257,8 @@ class MirrorStruct:
 			mir_rorp.flaglinked(Hardlink.get_link_index(mir_rorp))
 		elif mir_rorp.isreg():
 			expanded_index = cls.mirror_base.index + mir_rorp.index
-			mir_rorp.setfile(cls.rf_cache.get_fp(expanded_index))
+			file_fp = cls.rf_cache.get_fp(expanded_index, mir_rorp)
+			mir_rorp.setfile(hash.FileWrapper(file_fp))
 		mir_rorp.set_attached_filetype('snapshot')
 		return mir_rorp
 
@@ -349,34 +325,35 @@ class CachedRF:
 		s3 = "--------------------------"
 		return "\n".join((s1, s2, s3))
 
-	def get_rf(self, index):
-		"""Return RestoreFile of given index, or None"""
+	def get_rf(self, index, mir_rorp = None):
+		"""Get a RestoreFile for given index, or None"""
 		while 1:
 			if not self.rf_list:
-				if not self.add_rfs(index): return None
+				if not self.add_rfs(index, mir_rorp): return None
 			rf = self.rf_list[0]
 			if rf.index == index:
-				if Globals.process_uid != 0: self.perm_changer(index)
+				if Globals.process_uid != 0: self.perm_changer(index, mir_rorp)
 				return rf
 			elif rf.index > index:
 				# Try to add earlier indicies.  But if first is
 				# already from same directory, or we can't find any
 				# from that directory, then we know it can't be added.
 				if (index[:-1] == rf.index[:-1] or not
-					self.add_rfs(index)): return None
+					self.add_rfs(index, mir_rorp)): return None
 			else: del self.rf_list[0]
 
-	def get_fp(self, index):
+	def get_fp(self, index, mir_rorp):
 		"""Return the file object (for reading) of given index"""
-		rf = self.get_rf(index)
+		rf = longname.update_rf(self.get_rf(index, mir_rorp), mir_rorp,
+								self.root_rf.mirror_rp)
 		if not rf:
-			log.Log("""Error: Unable to retrieve data for file %s!
-The cause is probably data loss from the destination directory.""" %
-					(index and "/".join(index) or '.',), 2)
+			log.Log("Error: Unable to retrieve data for file %s!\nThe "
+					"cause is probably data loss from the backup repository."
+					% (index and "/".join(index) or '.',), 2)
 			return cStringIO.StringIO('')
-		return self.get_rf(index).get_restore_fp()
+		return rf.get_restore_fp()
 
-	def add_rfs(self, index):
+	def add_rfs(self, index, mir_rorp = None):
 		"""Given index, add the rfs in that same directory
 
 		Returns false if no rfs are available, which usually indicates
@@ -384,14 +361,13 @@ The cause is probably data loss from the destination directory.""" %
 
 		"""
 		if not index: return self.root_rf
+		if mir_rorp.has_alt_mirror_name(): return # longname alias separate
 		parent_index = index[:-1]
 		if Globals.process_uid != 0: self.perm_changer(parent_index)
 		temp_rf = RestoreFile(self.root_rf.mirror_rp.new_index(parent_index),
 							  self.root_rf.inc_rp.new_index(parent_index), [])
 		new_rfs = list(temp_rf.yield_sub_rfs())
-		if not new_rfs:
-			log.Log("Warning: No RFs added for index %s" % (index,), 2)
-			return 0
+		if not new_rfs: return 0
 		self.rf_list[0:0] = new_rfs
 		return 1
 
@@ -409,9 +385,6 @@ class RestoreFile:
 
 	"""
 	def __init__(self, mirror_rp, inc_rp, inc_list):
-		assert mirror_rp.index == inc_rp.index, \
-			   ("mirror and inc indicies don't match: %s %s" %
-				(mirror_rp.get_indexpath(), inc_rp.get_indexpath()))
 		self.index = mirror_rp.index
 		self.mirror_rp = mirror_rp
 		self.inc_rp, self.inc_list = inc_rp, inc_list
@@ -434,7 +407,8 @@ class RestoreFile:
 		"""
 		self.mirror_rp.inc_type = 'snapshot'
 		self.mirror_rp.inc_compressed = 0
-		if not self.inc_list or _rest_time >= _mirror_time:
+		if (not self.inc_list or
+			MirrorStruct._rest_time >= MirrorStruct._mirror_time):
 			self.relevant_incs = [self.mirror_rp]
 			return
 
@@ -461,7 +435,7 @@ class RestoreFile:
 		incpairs = []
 		for inc in self.inc_list:
 			time = inc.getinctime()
-			if time >= _rest_time: incpairs.append((time, inc))
+			if time >= MirrorStruct._rest_time: incpairs.append((time, inc))
 		incpairs.sort()
 		return [pair[1] for pair in incpairs]
 
@@ -484,6 +458,23 @@ class RestoreFile:
 
 	def get_restore_fp(self):
 		"""Return file object of restored data"""
+		def get_fp():
+			current_fp = self.get_first_fp()
+			for inc_diff in self.relevant_incs[1:]:
+				log.Log("Applying patch %s" % (inc_diff.get_indexpath(),), 7)
+				assert inc_diff.getinctype() == 'diff'
+				delta_fp = inc_diff.open("rb", inc_diff.isinccompressed())
+				new_fp = tempfile.TemporaryFile()
+				Rdiff.write_patched_fp(current_fp, delta_fp, new_fp)
+				new_fp.seek(0)
+				current_fp = new_fp
+			return current_fp
+
+		def error_handler(exc):
+			log.Log("Error reading %s, substituting empty file." %
+					(self.mirror_rp.path,), 2)
+			return cStringIO.StringIO('')
+
 		if not self.relevant_incs[-1].isreg():
 			log.Log("""Warning: Could not restore file %s!
 
@@ -492,18 +483,9 @@ constructed from existing increments because last increment had type
 %s.  Instead of the actual file's data, an empty length file will be
 created.  This error is probably caused by data loss in the
 rdiff-backup destination directory, or a bug in rdiff-backup""" %
-					(self.mirror_rp.path, self.relevant_incs[-1].lstat()), 2)
+	    (self.mirror_rp.get_indexpath(), self.relevant_incs[-1].lstat()), 2)
 			return cStringIO.StringIO('')
-		current_fp = self.get_first_fp()
-		for inc_diff in self.relevant_incs[1:]:
-			log.Log("Applying patch %s" % (inc_diff.get_indexpath(),), 7)
-			assert inc_diff.getinctype() == 'diff'
-			delta_fp = inc_diff.open("rb", inc_diff.isinccompressed())
-			new_fp = tempfile.TemporaryFile()
-			Rdiff.write_patched_fp(current_fp, delta_fp, new_fp)
-			new_fp.seek(0)
-			current_fp = new_fp
-		return current_fp
+		return robust.check_common_error(error_handler, get_fp)
 
 	def get_first_fp(self):
 		"""Return first file object from relevant inc list"""
@@ -521,13 +503,7 @@ rdiff-backup destination directory, or a bug in rdiff-backup""" %
 
 	def yield_sub_rfs(self):
 		"""Return RestoreFiles under current RestoreFile (which is dir)"""
-		if not self.mirror_rp.isdir() and not self.inc_rp.isdir():
-			log.Log("""Warning: directory %s seems to be missing from backup!
-
-This is probably due to files being deleted manually from the
-rdiff-backup destination directory.  In general you shouldn't do this,
-as data loss may result.\n""" % (self.mirror_rp.get_indexpath(),), 2)
-			return
+		if not self.mirror_rp.isdir() and not self.inc_rp.isdir(): return
 		if self.mirror_rp.isdir():
 			mirror_iter = self.yield_mirrorrps(self.mirror_rp)
 		else: mirror_iter = iter([])
@@ -633,15 +609,31 @@ class PatchITRB(rorpiter.ITRBranch):
 		self.patch_to_temp(rp, diff_rorp, tf)
 		rpath.rename(tf, rp)
 
+	def check_hash(self, copy_report, diff_rorp):
+		"""Check the hash in the copy_report with hash in diff_rorp"""
+		if not diff_rorp.isreg(): return
+		if not diff_rorp.has_sha1():
+			log.Log("Hash for %s missing, cannot check" %
+					(diff_rorp.get_indexpath()), 2)
+		elif copy_report.sha1_digest == diff_rorp.get_sha1():
+			log.Log("Hash %s of %s verified" %
+					(diff_rorp.get_sha1(), diff_rorp.get_indexpath()), 6)
+		else:
+			log.Log("Warning: Hash %s of %s\ndoesn't match recorded hash %s!"
+					% (copy_report.sha1_digest, diff_rorp.get_indexpath(),
+					   diff_rorp.get_sha1()), 2)
+
 	def patch_to_temp(self, basis_rp, diff_rorp, new):
 		"""Patch basis_rp, writing output in new, which doesn't exist yet"""
 		if diff_rorp.isflaglinked():
 			Hardlink.link_rp(diff_rorp, new, self.basis_root_rp)
-		elif diff_rorp.get_attached_filetype() == 'snapshot':
-			rpath.copy(diff_rorp, new)
+			return
+		if diff_rorp.get_attached_filetype() == 'snapshot':
+			copy_report = rpath.copy(diff_rorp, new)
 		else:
 			assert diff_rorp.get_attached_filetype() == 'diff'
-			Rdiff.patch_local(basis_rp, diff_rorp, new)
+			copy_report = Rdiff.patch_local(basis_rp, diff_rorp, new)
+		self.check_hash(copy_report, diff_rorp)
 		if new.lstat(): rpath.copy_attribs(diff_rorp, new)
 
 	def start_process(self, index, diff_rorp):
@@ -698,8 +690,9 @@ class PermissionChanger:
 		# order that need clearing
 		self.open_index_list = []
 
-	def __call__(self, index):
+	def __call__(self, index, mir_rorp = None):
 		"""Given rpath, change permissions up to and including index"""
+		if mir_rorp and mir_rorp.has_alt_mirror_name(): return
 		old_index = self.current_index
 		self.current_index = index
 		if not index or index <= old_index: return
@@ -718,7 +711,7 @@ class PermissionChanger:
 		"""Change permissions of directories between old_index and index"""
 		for rp in self.get_new_rp_list(old_index, index):
 			if ((rp.isreg() and not rp.readable()) or
-				(rp.isdir() and not rp.hasfullperms())):
+				(rp.isdir() and not (rp.executable() and rp.readable()))):
 				old_perms = rp.getperms()
 				self.open_index_list.insert(0, (rp.index, rp, old_perms))
 				if rp.isreg(): rp.chmod(0400 | old_perms)
@@ -743,3 +736,8 @@ class PermissionChanger:
 	def finish(self):
 		"""Restore any remaining rps"""
 		for index, rp, perms in self.open_index_list: rp.chmod(perms)
+
+
+import Globals, Time, Rdiff, Hardlink, selection, rpath, \
+	   log, robust, metadata, statistics, TempFile, hash, longname
+
