@@ -36,8 +36,12 @@ are dealing with are local or remote.
 """
 
 import os, stat, re, sys, shutil, gzip, socket, time, errno
-import Globals, Time, static, log, user_group
+import Globals, Time, static, log, user_group, C
 
+try:
+	import win32file, winnt
+except ImportError:
+	pass
 
 class SkipFileException(Exception):
 	"""Signal that the current file should be skipped but then continue
@@ -82,7 +86,12 @@ def move(rpin, rpout):
 		rpin.delete()
 
 def copy(rpin, rpout, compress = 0):
-	"""Copy RPath rpin to rpout.  Works for symlinks, dirs, etc."""
+	"""Copy RPath rpin to rpout.  Works for symlinks, dirs, etc.
+
+	Returns close value of input for regular file, which can be used
+	to pass hashes on.
+
+	"""
 	log.Log("Regular copying %s to %s" % (rpin.index, rpout.path), 6)
 	if not rpin.lstat():
 		if rpout.lstat(): rpout.delete()
@@ -93,9 +102,14 @@ def copy(rpin, rpout, compress = 0):
 			rpout.delete()   # easier to write than compare
 		else: return
 
-	if rpin.isreg(): copy_reg_file(rpin, rpout, compress)
+	if rpin.isreg(): return copy_reg_file(rpin, rpout, compress)
 	elif rpin.isdir(): rpout.mkdir()
-	elif rpin.issym(): rpout.symlink(rpin.readlink())
+	elif rpin.issym():
+		# some systems support permissions for symlinks, but
+		# only by setting at creation via the umask
+		if Globals.symlink_perms: orig_umask = os.umask(0777 & ~rpin.getperms())
+		rpout.symlink(rpin.readlink())
+		if Globals.symlink_perms: os.umask(orig_umask)	# restore previous umask
 	elif rpin.ischardev():
 		major, minor = rpin.getdevnums()
 		rpout.makedev("c", major, minor)
@@ -111,11 +125,21 @@ def copy_reg_file(rpin, rpout, compress = 0):
 	try:
 		if (rpout.conn is rpin.conn and
 			rpout.conn is not Globals.local_connection):
-			rpout.conn.rpath.copy_reg_file(rpin.path, rpout.path, compress)
+			v = rpout.conn.rpath.copy_reg_file(rpin.path, rpout.path, compress)
 			rpout.setdata()
-			return
+			return v
 	except AttributeError: pass
-	rpout.write_from_fileobj(rpin.open("rb"), compress = compress)
+	try:
+		return rpout.write_from_fileobj(rpin.open("rb"), compress = compress)
+	except IOError, e:
+		if (e.errno == errno.ERANGE):
+			log.Log.FatalError("'IOError - Result too large' while reading %s. "
+							   "If you are using a Mac, this is probably "
+							   "the result of HFS+ filesystem corruption. "
+							   "Please exclude this file from your backup "
+							   "before proceeding." % rpin.path)
+		else:
+			raise
 
 def cmp(rpin, rpout):
 	"""True if rpin has the same data as rpout
@@ -152,16 +176,20 @@ def copy_attribs(rpin, rpout):
 	"""
 	log.Log("Copying attributes from %s to %s" % (rpin.index, rpout.path), 7)
 	assert rpin.lstat() == rpout.lstat() or rpin.isspecial()
-	if Globals.change_ownership: rpout.chown(*user_group.map_rpath(rpin))
-	if rpin.issym(): return # symlinks don't have times or perms
-	if Globals.resource_forks_write and rpin.isreg():
-		rpout.write_resource_fork(rpin.get_resource_fork())
-	if Globals.carbonfile_write and rpin.isreg():
-		rpout.write_carbonfile(rpin.get_carbonfile())
+	if Globals.change_ownership:
+		rpout.chown(*rpout.conn.user_group.map_rpath(rpin))
 	if Globals.eas_write: rpout.write_ea(rpin.get_ea())
+	if rpin.issym(): return # symlinks don't have times or perms
+	if (Globals.resource_forks_write and rpin.isreg() and
+		rpin.has_resource_fork()):
+		rpout.write_resource_fork(rpin.get_resource_fork())
+	if (Globals.carbonfile_write and rpin.isreg() and
+		rpin.has_carbonfile()):
+		rpout.write_carbonfile(rpin.get_carbonfile())
 	rpout.chmod(rpin.getperms())
 	if Globals.acls_write: rpout.write_acl(rpin.get_acl())
 	if not rpin.isdev(): rpout.setmtime(rpin.getmtime())
+	if Globals.win_acls_write: rpout.write_win_acl(rpin.get_win_acl())
 
 def copy_attribs_inc(rpin, rpout):
 	"""Change file attributes of rpout to match rpin
@@ -174,12 +202,14 @@ def copy_attribs_inc(rpin, rpout):
 	log.Log("Copying inc attrs from %s to %s" % (rpin.index, rpout.path), 7)
 	check_for_files(rpin, rpout)
 	if Globals.change_ownership: apply(rpout.chown, rpin.getuidgid())
-	if rpin.issym(): return # symlinks don't have times or perms
-	if Globals.resource_forks_write and rpin.isreg() and rpout.isreg():
-		rpout.write_resource_fork(rpin.get_resource_fork())
-	if Globals.carbonfile_write and rpin.isreg() and rpout.isreg():
-		rpout.write_carbonfile(rpin.get_carbonfile())
 	if Globals.eas_write: rpout.write_ea(rpin.get_ea())
+	if rpin.issym(): return # symlinks don't have times or perms
+	if (Globals.resource_forks_write and rpin.isreg() and
+		rpin.has_resource_fork() and rpout.isreg()):
+		rpout.write_resource_fork(rpin.get_resource_fork())
+	if (Globals.carbonfile_write and rpin.isreg() and
+		rpin.has_carbonfile() and rpout.isreg()):
+		rpout.write_carbonfile(rpin.get_carbonfile())
 	if rpin.isdir() and not rpout.isdir():
 		rpout.chmod(rpin.getperms() & 0777)
 	else: rpout.chmod(rpin.getperms())
@@ -219,25 +249,101 @@ def rename(rp_source, rp_dest):
 	log.Log(lambda: "Renaming %s to %s" % (rp_source.path, rp_dest.path), 7)
 	if not rp_source.lstat(): rp_dest.delete()
 	else:
-		if rp_dest.lstat() and rp_source.getinode() == rp_dest.getinode():
+		if rp_dest.lstat() and rp_source.getinode() == rp_dest.getinode() and \
+				rp_source.getinode() != 0:
 			log.Log("Warning: Attempt to rename over same inode: %s to %s"
 					% (rp_source.path, rp_dest.path), 2)
 			# You can't rename one hard linked file over another
 			rp_source.delete()
-		else: rp_source.conn.os.rename(rp_source.path, rp_dest.path)
+		else:
+			try:
+			    rp_source.conn.os.rename(rp_source.path, rp_dest.path)
+			except OSError, error:
+				# XXX errno.EINVAL and len(rp_dest.path) >= 260 indicates
+				# pathname too long on Windows
+				if error.errno != errno.EEXIST:
+					log.Log("OSError while renaming %s to %s"
+							% (rp_source.path, rp_dest.path), 1)
+					raise
+
+				# On Windows, files can't be renamed on top of an existing file
+				rp_source.conn.os.chmod(rp_dest.path, 0700)
+				rp_source.conn.os.unlink(rp_dest.path)
+				rp_source.conn.os.rename(rp_source.path, rp_dest.path)
+			    
 		rp_dest.data = rp_source.data
 		rp_source.data = {'type': None}
 
-def tupled_lstat(filename):
-	"""Like os.lstat, but return only a tuple, or None if os.error
-
-	Later versions of os.lstat return a special lstat object,
-	which can confuse the pickler and cause errors in remote
-	operations.  This has been fixed in Python 2.2.1.
+def make_file_dict(filename):
+	"""Generate the data dictionary for the given RPath
 
+	This is a global function so that os.name can be called locally,
+	thus avoiding network lag and so that we only need to send the
+	filename over the network, thus avoiding the need to pickle an
+	(incomplete) rpath object.
 	"""
-	try: return tuple(os.lstat(filename))
-	except os.error: return None
+	if os.name != 'nt':
+		try:
+			return C.make_file_dict(filename)
+		except OSError, error:
+			# Unicode filenames should be process by the Python version 
+			if error.errno != errno.EILSEQ and error.errno != errno.EINVAL:
+				raise
+
+	return make_file_dict_python(filename)
+
+def make_file_dict_python(filename):
+	"""Create the data dictionary using a Python call to os.lstat
+	
+	We do this on Windows since Python's implementation is much better
+	than the one in cmodule.c    Eventually, we will move to using
+	this on all platforms since CPUs have gotten much faster than
+	they were when it was necessary to write cmodule.c
+	"""
+	try:
+		statblock = os.lstat(filename)
+	except os.error:
+		return {'type':None}
+	data = {}
+	mode = statblock[stat.ST_MODE]
+
+	if stat.S_ISREG(mode): type = 'reg'
+	elif stat.S_ISDIR(mode): type = 'dir'
+	elif stat.S_ISCHR(mode):
+		type = 'dev'
+		s = statblock.st_rdev
+		data['devnums'] = ('c',) + (s >> 8, s & 0xff)
+	elif stat.S_ISBLK(mode):
+		type = 'dev'
+		s = statblock.st_rdev
+		data['devnums'] = ('b',) + (s >> 8, s & 0xff)
+	elif stat.S_ISFIFO(mode): type = 'fifo'
+	elif stat.S_ISLNK(mode):
+		type = 'sym'
+		data['linkname'] = os.readlink(filename)
+	elif stat.S_ISSOCK(mode): type = 'sock'
+	else: raise C.UnknownFileError(filename)
+	data['type'] = type
+	data['size'] = statblock[stat.ST_SIZE]
+	data['perms'] = stat.S_IMODE(mode)
+	data['uid'] = statblock[stat.ST_UID]
+	data['gid'] = statblock[stat.ST_GID]
+	data['inode'] = statblock[stat.ST_INO]
+	data['devloc'] = statblock[stat.ST_DEV]
+	data['nlink'] = statblock[stat.ST_NLINK]
+
+	if os.name == 'nt':
+		attribs = win32file.GetFileAttributes(filename)
+		if attribs & winnt.FILE_ATTRIBUTE_REPARSE_POINT:
+			data['type'] = 'sym'
+			data['linkname'] = None
+
+	if not (type == 'sym' or type == 'dev'):
+		# mtimes on symlinks and dev files don't work consistently
+		data['mtime'] = long(statblock[stat.ST_MTIME])
+		data['atime'] = long(statblock[stat.ST_ATIME])
+		data['ctime'] = long(statblock[stat.ST_CTIME])
+	return data
 
 def make_socket_local(rpath):
 	"""Make a local socket at the given path
@@ -255,13 +361,41 @@ def make_socket_local(rpath):
 def gzip_open_local_read(rpath):
 	"""Return open GzipFile.  See security note directly above"""
 	assert rpath.conn is Globals.local_connection
-	return gzip.GzipFile(rpath.path, "rb")
+	return GzipFile(rpath.path, "rb")
 
 def open_local_read(rpath):
 	"""Return open file (provided for security reasons)"""
 	assert rpath.conn is Globals.local_connection
 	return open(rpath.path, "rb")
 
+def get_incfile_info(basename):
+	"""Returns None or tuple of 
+	(is_compressed, timestr, type, and basename)"""
+	dotsplit = basename.split(".")
+	if dotsplit[-1] == "gz":
+		compressed = 1
+		if len(dotsplit) < 4: return None
+		timestring, ext = dotsplit[-3:-1]
+	else:
+		compressed = None
+		if len(dotsplit) < 3: return None
+		timestring, ext = dotsplit[-2:]
+	if Time.stringtotime(timestring) is None: return None
+	if not (ext == "snapshot" or ext == "dir" or
+			ext == "missing" or ext == "diff" or ext == "data"):
+		return None
+	if compressed: basestr = ".".join(dotsplit[:-3])
+	else: basestr = ".".join(dotsplit[:-2])
+	return (compressed, timestring, ext, basestr)
+
+def delete_dir_no_files(rp):
+	"""Deletes the directory at rp.path if empty. Raises if the
+	directory contains files."""
+	assert rp.isdir()
+	if rp.contains_files():
+		raise RPathException("Directory contains files.")
+	rp.delete()
+
 
 class RORPath:
 	"""Read Only RPath - carry information about a path
@@ -303,6 +437,7 @@ class RORPath:
 			elif key == 'size' and not self.isreg(): pass
 			elif key == 'ea' and not Globals.eas_active: pass
 			elif key == 'acl' and not Globals.acls_active: pass
+			elif key == 'win_acl' and not Globals.win_acls_active: pass
 			elif key == 'carbonfile' and not Globals.carbonfile_active: pass
 			elif key == 'resourcefork' and not Globals.resource_forks_active:
 				pass
@@ -343,22 +478,29 @@ class RORPath:
 			elif key == 'inode': pass
 			elif key == 'ea' and not Globals.eas_write: pass
 			elif key == 'acl' and not Globals.acls_write: pass
+			elif key == 'win_acl' and not Globals.win_acls_write: pass
 			elif key == 'carbonfile' and not Globals.carbonfile_write: pass
 			elif key == 'resourcefork' and not Globals.resource_forks_write:
 				pass
+			elif key == 'sha1': pass # one or other may not have set
+			elif key == 'mirrorname' or key == 'incname': pass
 			elif (not other.data.has_key(key) or
-				  self.data[key] != other.data[key]): return 0
+				  self.data[key] != other.data[key]):
+				return 0
 
 		if self.lstat() and not self.issym() and Globals.change_ownership:
 			# Now compare ownership.  Symlinks don't have ownership
-			if user_group.map_rpath(self) != other.getuidgid(): return 0
+			try:
+				if user_group.map_rpath(self) != other.getuidgid(): return 0
+			except KeyError:
+				return 0 # uid/gid might be missing if metadata file is corrupt
 
 		return 1
 
 	def equal_verbose(self, other, check_index = 1,
 					  compare_inodes = 0, compare_ownership = 0,
-					  compare_acls = 0, compare_eas = 0, compare_size = 1,
-					  compare_type = 1, verbosity = 2):
+					  compare_acls = 0, compare_eas = 0, compare_win_acls = 0,
+					  compare_size = 1, compare_type = 1, verbosity = 2):
 		"""Like __eq__, but log more information.  Useful when testing"""
 		if check_index and self.index != other.index:
 			log.Log("Index %s != index %s" % (self.index, other.index),
@@ -379,6 +521,7 @@ class RORPath:
 				pass
 			elif key == 'ea' and not compare_eas: pass
 			elif key == 'acl' and not compare_acls: pass
+			elif key == 'win_acl' and not compare_win_acls: pass
 			elif (not other.data.has_key(key) or
 				  self.data[key] != other.data[key]):
 				if not other.data.has_key(key):
@@ -396,7 +539,8 @@ class RORPath:
 		return self.equal_verbose(other,
 								  compare_inodes = compare_inodes,
 								  compare_eas = Globals.eas_active,
-								  compare_acls = Globals.acls_active)
+								  compare_acls = Globals.acls_active,
+								  compare_win_acls = Globals.win_acls_active)
 							 
 	def __ne__(self, other): return not self.__eq__(other)
 
@@ -477,7 +621,8 @@ class RORPath:
 
 	def getperms(self):
 		"""Return permission block of file"""
-		return self.data['perms']
+		if self.data.has_key('perms'): return self.data['perms']
+		else: return 0
 
 	def getuname(self):
 		"""Return username that owns the file"""
@@ -603,7 +748,10 @@ class RORPath:
 
 	def get_acl(self):
 		"""Return access control list object from dictionary"""
-		return self.data['acl']
+		try: return self.data['acl']
+		except KeyError:
+			acl = self.data['acl'] = get_blank_acl(self.index)
+			return acl
 
 	def set_ea(self, ea):
 		"""Record extended attributes in dictionary.  Does not write"""
@@ -611,7 +759,10 @@ class RORPath:
 
 	def get_ea(self):
 		"""Return extended attributes object"""
-		return self.data['ea']
+		try: return self.data['ea']
+		except KeyError:
+			ea = self.data['ea'] = get_blank_ea(self.index)
+			return ea
 
 	def has_carbonfile(self):
 		"""True if rpath has a carbonfile parameter"""
@@ -637,6 +788,65 @@ class RORPath:
 		"""Record resource fork in dictionary.  Does not write"""
 		self.data['resourcefork'] = rfork
 
+	def set_win_acl(self, acl):
+		"""Record Windows access control list in dictionary. Does not write"""
+		self.data['win_acl'] = acl
+
+	def get_win_acl(self):
+		"""Return access control list object from dictionary"""
+		try: return self.data['win_acl']
+		except KeyError:
+			acl = self.data['win_acl'] = get_blank_win_acl(self.index)
+			return acl
+
+	def has_alt_mirror_name(self):
+		"""True if rorp has an alternate mirror name specified"""
+		return self.data.has_key('mirrorname')
+
+	def get_alt_mirror_name(self):
+		"""Return alternate mirror name (for long filenames)"""
+		return self.data['mirrorname']
+
+	def set_alt_mirror_name(self, filename):
+		"""Set alternate mirror name to filename
+
+		Instead of writing to the traditional mirror file, store
+		mirror information in filename in the long filename
+		directory.
+
+		"""
+		self.data['mirrorname'] = filename
+
+	def has_alt_inc_name(self):
+		"""True if rorp has an alternate increment base specified"""
+		return self.data.has_key('incname')
+
+	def get_alt_inc_name(self):
+		"""Return alternate increment base (used for long name support)"""
+		return self.data['incname']
+
+	def set_alt_inc_name(self, name):
+		"""Set alternate increment name to name
+
+		If set, increments will be in the long name directory with
+		name as their base.  If the alt mirror name is set, this
+		should be set to the same.
+
+		"""
+		self.data['incname'] = name
+
+	def has_sha1(self):
+		"""True iff self has its sha1 digest set"""
+		return self.data.has_key('sha1')
+
+	def get_sha1(self):
+		"""Return sha1 digest.  Causes exception unless set_sha1 first"""
+		return self.data['sha1']
+
+	def set_sha1(self, digest):
+		"""Set sha1 hash (should be in hexdecimal)"""
+		self.data['sha1'] = digest
+
 
 class RPath(RORPath):
 	"""Remote Path class - wrapper around a possibly non-local pathname
@@ -680,62 +890,24 @@ class RPath(RORPath):
 	def __getstate__(self):
 		"""Return picklable state
 
-		The connection must be local because we can't pickle a
-		connection.  Data and any attached file also won't be saved.
+		The rpath's connection will be encoded as its conn_number.  It
+		and the other information is put in a tuple. Data and any attached
+		file won't be saved.
 
 		"""
-		assert self.conn is Globals.local_connection
-		return (self.index, self.base, self.data)
+		return (self.conn.conn_number, self.base, self.index, self.data)
 
 	def __setstate__(self, rpath_state):
 		"""Reproduce RPath from __getstate__ output"""
-		self.conn = Globals.local_connection
-		self.index, self.base, self.data = rpath_state
+		conn_number, self.base, self.index, self.data = rpath_state
+		self.conn = Globals.connection_dict[conn_number]
 		self.path = "/".join((self.base,) + self.index)
 
 	def setdata(self):
-		"""Set data dictionary using C extension"""
-		self.data = self.conn.C.make_file_dict(self.path)
+		"""Set data dictionary using the wrapper"""
+		self.data = self.conn.rpath.make_file_dict(self.path)
 		if self.lstat(): self.conn.rpath.setdata_local(self)
 
-	def make_file_dict_old(self):
-		"""Create the data dictionary"""
-		statblock = self.conn.rpath.tupled_lstat(self.path)
-		if statblock is None:
-			return {'type':None}
-		data = {}
-		mode = statblock[stat.ST_MODE]
-
-		if stat.S_ISREG(mode): type = 'reg'
-		elif stat.S_ISDIR(mode): type = 'dir'
-		elif stat.S_ISCHR(mode):
-			type = 'dev'
-			data['devnums'] = ('c',) + self._getdevnums()
-		elif stat.S_ISBLK(mode):
-			type = 'dev'
-			data['devnums'] = ('b',) + self._getdevnums()
-		elif stat.S_ISFIFO(mode): type = 'fifo'
-		elif stat.S_ISLNK(mode):
-			type = 'sym'
-			data['linkname'] = self.conn.os.readlink(self.path)
-		elif stat.S_ISSOCK(mode): type = 'sock'
-		else: raise C.UnknownFileError(self.path)
-		data['type'] = type
-		data['size'] = statblock[stat.ST_SIZE]
-		data['perms'] = stat.S_IMODE(mode)
-		data['uid'] = statblock[stat.ST_UID]
-		data['gid'] = statblock[stat.ST_GID]
-		data['inode'] = statblock[stat.ST_INO]
-		data['devloc'] = statblock[stat.ST_DEV]
-		data['nlink'] = statblock[stat.ST_NLINK]
-
-		if not (type == 'sym' or type == 'dev'):
-			# mtimes on symlinks and dev files don't work consistently
-			data['mtime'] = long(statblock[stat.ST_MTIME])
-			data['atime'] = long(statblock[stat.ST_ATIME])
-			data['ctime'] = long(statblock[stat.ST_CTIME])
-		return data
-
 	def check_consistency(self):
 		"""Raise an error if consistency of rp broken
 
@@ -749,14 +921,22 @@ class RPath(RORPath):
 			   "\nName: %s\nOld: %s --> New: %s\n" % \
 			   (self.path, temptype, self.data['type'])
 
-	def _getdevnums(self):
-		"""Return tuple for special file (major, minor)"""
-		s = self.conn.reval("lambda path: os.lstat(path).st_rdev", self.path)
-		return (s >> 8, s & 0xff)
-
-	def chmod(self, permissions):
+	def chmod(self, permissions, loglevel = 2):
 		"""Wrapper around os.chmod"""
-		self.conn.os.chmod(self.path, permissions & Globals.permission_mask)
+		try:
+			self.conn.os.chmod(self.path, permissions & Globals.permission_mask)
+		except OSError, e:
+			if e.strerror == "Inappropriate file type or format" \
+					and not self.isdir():
+				# Some systems throw this error if try to set sticky bit
+				# on a non-directory. Remove sticky bit and try again.
+				log.Log("Warning: Unable to set permissions of %s to %o - "
+						"trying again without sticky bit (%o)" % (self.path, 
+						permissions, permissions & 06777), loglevel)
+				self.conn.os.chmod(self.path, permissions
+											  & 06777 & Globals.permission_mask)
+			else:
+				raise
 		self.data['perms'] = permissions
 
 	def settime(self, accesstime, modtime):
@@ -774,10 +954,17 @@ class RPath(RORPath):
 	def setmtime(self, modtime):
 		"""Set only modtime (access time to present)"""
 		log.Log(lambda: "Setting time of %s to %d" % (self.path, modtime), 7)
+		if modtime < 0: log.Log("Warning: modification time of %s is"
+								"before 1970" % self.path, 2)
 		try: self.conn.os.utime(self.path, (long(time.time()), modtime))
 		except OverflowError:
 			log.Log("Cannot change mtime of %s to %s - problem is probably"
 					"64->32bit conversion" % (self.path, modtime), 2)
+		except OSError:
+			# It's not possible to set a modification time for
+			# directories on Windows.
+		    if self.conn.os.name != 'nt' or not self.isdir():
+		        raise
 		else: self.data['mtime'] = modtime
 
 	def chown(self, uid, gid):
@@ -861,24 +1048,54 @@ class RPath(RORPath):
 		
 	def isowner(self):
 		"""Return true if current process is owner of rp or root"""
-		uid = self.conn.os.getuid()
-		return uid == 0 or uid == self.data['uid']
+		try:
+		    uid = self.conn.os.getuid()
+		except AttributeError:
+		    return True # Windows doesn't have getuid(), so hope for the best
+		return uid == 0 or \
+			   (self.data.has_key('uid') and uid == self.data['uid'])
 
 	def isgroup(self):
 		"""Return true if process has group of rp"""
-		return self.data['gid'] in self.conn.Globals.get('process_groups')
+		return (self.data.has_key('gid') and \
+				self.data['gid'] in self.conn.Globals.get('process_groups'))
 
 	def delete(self):
 		"""Delete file at self.path.  Recursively deletes directories."""
 		log.Log("Deleting %s" % self.path, 7)
 		if self.isdir():
 			try: self.rmdir()
-			except EnvironmentError:
+			except os.error:
 				if Globals.fsync_directories: self.fsync()
 				self.conn.shutil.rmtree(self.path)
-		else: self.conn.os.unlink(self.path)
+		else:
+			try: self.conn.os.unlink(self.path)
+			except OSError, error:
+				if error.errno in (errno.EPERM, errno.EACCES):
+					# On Windows, read-only files cannot be deleted.
+					# Remove the read-only attribute and try again.
+					self.chmod(0700)
+					self.conn.os.unlink(self.path)
+				else:
+					raise
+
 		self.setdata()
 
+	def contains_files(self):
+		"""Returns true if self (or subdir) contains any regular files."""
+		log.Log("Determining if directory contains files: %s" % self.path, 7)
+		if not self.isdir():
+			return False
+		dir_entries = self.listdir()
+		for entry in dir_entries:
+			child_rp = self.append(entry)
+			if not child_rp.isdir():
+				return True
+			else:
+				if child_rp.contains_files():
+					return True
+		return False
+
 	def quote(self):
 		"""Return quoted self.path for use with os.system()"""
 		return '"%s"' % self.regex_chars_to_quote.sub(
@@ -953,13 +1170,13 @@ class RPath(RORPath):
 
 		"""
 		if self.conn is Globals.local_connection:
-			if compress: return gzip.GzipFile(self.path, mode)
+			if compress: return GzipFile(self.path, mode)
 			else: return open(self.path, mode)
 
 		if compress:
 			if mode == "r" or mode == "rb":
 				return self.conn.rpath.gzip_open_local_read(self)
-			else: return self.conn.gzip.GzipFile(self.path, mode)
+			else: return self.conn.rpath.GzipFile(self.path, mode)
 		else:
 			if mode == "r" or mode == "rb":
 				return self.conn.rpath.open_local_read(self)
@@ -969,16 +1186,16 @@ class RPath(RORPath):
 		"""Reads fp and writes to self.path.  Closes both when done
 
 		If compress is true, fp will be gzip compressed before being
-		written to self.
+		written to self.  Returns closing value of fp.
 
 		"""
 		log.Log("Writing file object to " + self.path, 7)
 		assert not self.lstat(), "File %s already exists" % self.path
 		outfp = self.open("wb", compress = compress)
 		copyfileobj(fp, outfp)
-		if fp.close() or outfp.close():
-			raise RPathException("Error closing file")
+		if outfp.close(): raise RPathException("Error closing file")
 		self.setdata()
+		return fp.close()
 
 	def write_string(self, s, compress = None):
 		"""Write string s into rpath"""
@@ -994,26 +1211,17 @@ class RPath(RORPath):
 		Also sets various inc information used by the *inc* functions.
 
 		"""
-		if self.index: dotsplit = self.index[-1].split(".")
-		else: dotsplit = self.base.split(".")
-		if dotsplit[-1] == "gz":
-			compressed = 1
-			if len(dotsplit) < 4: return None
-			timestring, ext = dotsplit[-3:-1]
+		if self.index: basename = self.index[-1]
+		else: basename = self.base
+
+		inc_info = get_incfile_info(basename)
+
+		if inc_info:
+			self.inc_compressed, self.inc_timestr, \
+				self.inc_type, self.inc_basestr = inc_info
+			return 1
 		else:
-			compressed = None
-			if len(dotsplit) < 3: return None
-			timestring, ext = dotsplit[-2:]
-		if Time.stringtotime(timestring) is None: return None
-		if not (ext == "snapshot" or ext == "dir" or
-				ext == "missing" or ext == "diff" or ext == "data"):
 			return None
-		self.inc_timestr = timestring
-		self.inc_compressed = compressed
-		self.inc_type = ext
-		if compressed: self.inc_basestr = ".".join(dotsplit[:-3])
-		else: self.inc_basestr = ".".join(dotsplit[:-2])
-		return 1
 
 	def isinccompressed(self):
 		"""Return true if inc file is compressed"""
@@ -1068,35 +1276,47 @@ class RPath(RORPath):
 		if not fp: self.conn.rpath.RPath.fsync_local(self)
 		else: os.fsync(fp.fileno())
 
-	def fsync_local(self):
-		"""fsync current file, run locally"""
+	def fsync_local(self, thunk = None):
+		"""fsync current file, run locally
+
+		If thunk is given, run it before syncing but after gathering
+		the file's file descriptor.
+
+		"""
 		assert self.conn is Globals.local_connection
-		fd = os.open(self.path, os.O_RDONLY)
-		os.fsync(fd)
-		os.close(fd)
+		try:
+			fd = os.open(self.path, os.O_RDONLY)
+			os.fsync(fd)
+			os.close(fd)
+		except OSError, e:
+			if locals().has_key('fd'): os.close(fd)
+			if (e.errno not in (errno.EPERM, errno.EACCES, errno.EBADF)) \
+				or self.isdir(): raise
+
+			# Maybe the system doesn't like read-only fsyncing.
+			# However, to open RDWR, we may need to alter permissions
+			# temporarily.
+			if self.hasfullperms(): oldperms = None
+			else:
+				oldperms = self.getperms()
+				if not oldperms: # self.data['perms'] is probably out of sync
+					self.setdata()
+					oldperms = self.getperms()
+				self.chmod(0700)
+			fd = os.open(self.path, os.O_RDWR)
+			if oldperms is not None: self.chmod(oldperms)
+			if thunk: thunk()
+			os.fsync(fd) # Sync after we switch back permissions!
+			os.close(fd)
 
 	def fsync_with_dir(self, fp = None):
 		"""fsync self and directory self is under"""
 		self.fsync(fp)
 		if Globals.fsync_directories: self.get_parent_rp().fsync()
 
-	def sync_delete(self):
-		"""Delete self with sync to guarantee completion
-
-		On some filesystems (like linux's ext2), we must sync both the
-		file and the directory to make sure.
-
-		"""
-		if self.lstat() and not self.issym():
-			fp = self.open("rb")
-			self.delete()
-			os.fsync(fp.fileno())
-		assert not fp.close()
-		if Globals.fsync_directories: self.get_parent_rp().fsync()
-
-	def get_data(self):
+	def get_data(self, compressed = None):
 		"""Open file as a regular file, read data, close, return data"""
-		fp = self.open("rb")
+		fp = self.open("rb", compressed)
 		s = fp.read()
 		assert not fp.close()
 		return s
@@ -1127,38 +1347,13 @@ class RPath(RORPath):
 		ea.write_to_rp(self)
 		self.data['ea'] = ea
 
-	def get_carbonfile(self):
-		"""Return resource fork data, loading from filesystem if
-		necessary."""
-		from Carbon.File import FSSpec
-		import MacOS
-		try: return self.data['cfile']
-		except KeyError: pass
-
-		try:
-			fsobj = FSSpec(self.path)
-			finderinfo = fsobj.FSpGetFInfo()
-			cfile = {'creator': finderinfo.Creator,
-					 'type': finderinfo.Type,
-					 'location': finderinfo.Location,
-					 'flags': finderinfo.Flags}
-			self.data['carbonfile'] = cfile
-			return cfile
-		except MacOS.Error:
-			log.Log("Cannot read carbonfile information from %s" %
-					(self.path,), 2)
-			self.data['carbonfile'] = None 
-			return self.data['carbonfile']
-
 	def write_carbonfile(self, cfile):
 		"""Write new carbon data to self."""
-		if not cfile:
-			# This should be made cleaner---if you know Mac OS X tell
-			# me what could cause an error in get_carbonfile above
-			return
-			
+		if not cfile: return
 		log.Log("Writing carbon data to %s" % (self.index,), 7)
 		from Carbon.File import FSSpec
+		from Carbon.File import FSRef
+		import Carbon.Files
 		import MacOS
 		fsobj = FSSpec(self.path)
 		finderinfo = fsobj.FSpGetFInfo()
@@ -1167,7 +1362,16 @@ class RPath(RORPath):
 		finderinfo.Location = cfile['location']
 		finderinfo.Flags = cfile['flags']
 		fsobj.FSpSetFInfo(finderinfo)
-		self.set_carbonfile(cfile)
+
+		"""Write Creation Date to self (if stored in metadata)."""
+		try:
+			cdate = cfile['createDate']
+			fsref = FSRef(fsobj)
+			cataloginfo, d1, d2, d3 = fsref.FSGetCatalogInfo(Carbon.Files.kFSCatInfoCreateDate)
+			cataloginfo.createDate = (0, cdate, 0)
+			fsref.FSSetCatalogInfo(Carbon.Files.kFSCatInfoCreateDate, cataloginfo)
+			self.set_carbonfile(cfile)
+		except KeyError: self.set_carbonfile(cfile)
 
 	def get_resource_fork(self):
 		"""Return resource fork data, setting if necessary"""
@@ -1191,6 +1395,16 @@ class RPath(RORPath):
 		assert not fp.close()
 		self.set_resource_fork(rfork_data)
 
+	def get_win_acl(self):
+		"""Return Windows access control list, setting if necessary"""
+		try: acl = self.data['win_acl']
+		except KeyError: acl = self.data['win_acl'] = win_acl_get(self)
+		return acl
+
+	def write_win_acl(self, acl):
+		"""Change access control list of rp"""
+		write_win_acl(self, acl)
+		self.data['win_acl'] = acl
 
 class RPathFileHook:
 	"""Look like a file, but add closing hook"""
@@ -1208,6 +1422,67 @@ class RPathFileHook:
 		return result
 
 
+class GzipFile(gzip.GzipFile):
+	"""Like gzip.GzipFile, except remove destructor
+
+	The default GzipFile's destructor prints out some messy error
+	messages.  Use this class instead to clean those up.
+
+	"""
+	def __del__(self): pass
+	def __getattr__(self, name):
+		if name == 'fileno': return self.fileobj.fileno
+		else: raise AttributeError(name)
+
+
+class MaybeGzip:
+	"""Represent a file object that may or may not be compressed
+
+	We don't want to compress 0 length files.  This class lets us
+	delay the opening of the file until either the first write (so we
+	know it has data and should be compressed), or close (when there's
+	no data).
+
+	"""
+	def __init__(self, base_rp, callback = None):
+		"""Return file-like object with filename based on base_rp"""
+		assert not base_rp.lstat(), base_rp
+		self.base_rp = base_rp
+		# callback will be called with final write rp as only argument
+		self.callback = callback
+		self.fileobj = None # Will be None unless data gets written
+		self.closed = 0
+
+	def __getattr__(self, name):
+		if name == 'fileno': return self.fileobj.fileno
+		else: raise AttributeError(name)
+
+	def get_gzipped_rp(self):
+		"""Return gzipped rp by adding .gz to base_rp"""
+		if self.base_rp.index:
+			newind = self.base_rp.index[:-1] + (self.base_rp.index[-1]+'.gz',)
+			return self.base_rp.new_index(newind)
+		else: return self.base_rp.append_path('.gz')
+
+	def write(self, buf):
+		"""Write buf to fileobj"""
+		if self.fileobj: return self.fileobj.write(buf)
+		if not buf: return
+
+		new_rp = self.get_gzipped_rp()
+		if self.callback: self.callback(new_rp)
+		self.fileobj = new_rp.open("wb", compress = 1)
+		return self.fileobj.write(buf)
+
+	def close(self):
+		"""Close related fileobj, pass return value"""
+		if self.closed: return None
+		self.closed = 1
+		if self.fileobj: return self.fileobj.close()
+		if self.callback: self.callback(self.base_rp)
+		self.base_rp.touch()
+
+
 def setdata_local(rpath):
 	"""Set eas/acls, uid/gid, resource fork in data dictionary
 
@@ -1220,11 +1495,46 @@ def setdata_local(rpath):
 	rpath.data['gname'] = user_group.gid2gname(rpath.data['gid'])
 	if Globals.eas_conn: rpath.data['ea'] = ea_get(rpath)
 	if Globals.acls_conn: rpath.data['acl'] = acl_get(rpath)
+	if Globals.win_acls_conn:
+		rpath.data['win_acl'] = win_acl_get(rpath)
 	if Globals.resource_forks_conn and rpath.isreg():
 		rpath.get_resource_fork()
-	if Globals.carbonfile_conn and rpath.isreg(): rpath.get_carbonfile()
+	if Globals.carbonfile_conn and rpath.isreg():
+		rpath.data['carbonfile'] = carbonfile_get(rpath)
 
-# These two are overwritten by the eas_acls.py module.  We can't
+def carbonfile_get(rpath):
+	"""Return carbonfile value for local rpath"""
+	# Note, after we drop support for Mac OS X 10.0 - 10.3, it will no longer
+	# be necessary to read the finderinfo struct since it is a strict subset
+	# of the data stored in the com.apple.FinderInfo extended attribute
+	# introduced in 10.4. Indeed, FSpGetFInfo() is deprecated on 10.4.
+	from Carbon.File import FSSpec
+	from Carbon.File import FSRef
+	import Carbon.Files
+	import MacOS
+	try:
+		fsobj = FSSpec(rpath.path)
+		finderinfo = fsobj.FSpGetFInfo()
+		cataloginfo, d1, d2, d3 = FSRef(fsobj).FSGetCatalogInfo(Carbon.Files.kFSCatInfoCreateDate)
+		cfile = {'creator': finderinfo.Creator,
+				 'type': finderinfo.Type,
+				 'location': finderinfo.Location,
+				 'flags': finderinfo.Flags,
+				 'createDate': cataloginfo.createDate[1]}
+		return cfile
+	except MacOS.Error:
+		log.Log("Cannot read carbonfile information from %s" %
+				(rpath.path,), 2)
+		return None
+
+
+# These functions are overwritten by the eas_acls.py module.  We can't
 # import that module directly because of circular dependency problems.
 def acl_get(rp): assert 0
+def get_blank_acl(index): assert 0
 def ea_get(rp): assert 0
+def get_blank_ea(index): assert 0
+
+def win_acl_get(rp): assert 0
+def write_win_acl(rp): assert 0
+def get_blank_win_acl(): assert 0
